package edu.wpi.cs.jsh;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import sun.awt.X11.ToplevelStateListener;

import java.io.PrintStream;

/**
 * This factory allows for getting and creation of new streams, getting access to the top level streams, and controlling
 * output.
 */
public class JshStreamController {

    private static final Object _lock = new Object();
    private static final int DEFAULT_MAX_BUFFER = 10;
    private static JshStreamController instance;
    private static boolean initialized;

    /**
     * Initializes the JshStream api with the given output sources for stdout and stderr, and the given max buffer size.
     * If you want to just use the system default stdout and stderr, call {@link #initialize()} instead. If you want to
     * control either the buffer size or the streams individually, call the appropriate initialize method. The
     * controller cannot be reinitialized after it has been initialized once, and it will throw an exception if such
     * an instance occurs
     *
     * @param stdOut The printstream to write stdout to
     * @param stdErr The printstream to write stderr to
     * @throws JshException If the controller is already initialized, or if there is an error setting up the streams
     */
    public static void initialize(PrintStream stdOut, PrintStream stdErr, int maxBuffer) throws JshException {
        // Argument guards
        if (stdOut == null) {
            throw new JshException("Given StdOut source was null!");
        }
        if (stdErr == null) {
            throw new JshException("Given StdErr source was null!");
        }

        synchronized (_lock) {
            if (initialized) {
                throw new JshException("JshStream is already initialized! Cannot reinitialize in the middle of output.");
            }

            instance = new JshStreamController(stdOut, stdErr, maxBuffer);
            initialized = true;
        }
    }

    /**
     * Initializes the JshStream api with the default sysout and syserr, and a given buffer size. If the api has already
     * been initialized, then it will throw a JshException
     *
     * @param maxBuffer The buffer size to use
     * @throws JshException If the controller has already been initialized, or if there is an error setting up the stream
     */
    public static void initialize(int maxBuffer) throws JshException {
        initialize(System.out, System.err, maxBuffer);
    }

    /**
     * Initializes the JshStream api with the given sysout and syserr, and a default buffer size. If the api has already
     * been initialized, then it will throw a JshException
     *
     * @param sysOut The printstream to write stdout to
     * @param sysErr The printstream to write stderr to
     * @throws JshException If the controller has already been initialized, or if there is an error setting up the stream
     */
    public static void initialized(PrintStream sysOut, PrintStream sysErr) throws JshException {
        initialize(sysOut, sysErr, DEFAULT_MAX_BUFFER);
    }

    /**
     * Initializes the JshStream api with the default sysout and syserr sources. If the api has already been initialized,
     * then it will throw a JshException.
     *
     * @throws JshException If the controller has already been initialized, or if there is an error setting up the streams
     */
    public static void initialize() throws JshException {
        initialize(System.out, System.err, DEFAULT_MAX_BUFFER);
    }

    public static JshStreamController getInstance() throws JshException {
        if (!initialized) {
            throw new JshException("The JshController has not been set up! Please call one of the initialize methods.");
        }
        return instance;
    }

    /**
     * Resets the JshStreamController. After calling this, initialize must be called again.
     */
    public static void reset() {
        synchronized (_lock) {
            initialized = false;
            instance = null;
        }
    }

    public final Stream StdOutStream;
    public final Stream StdErrStream;
    private final PrintStream m_stdOutPrinter;
    private final PrintStream m_stdErrPrinter;
    private final int m_maxBuffer;
    private final Gson m_json;
    private int m_bufferCount;

    private JshStreamController(PrintStream stdOutPrinter, PrintStream stdErrPrinter, int maxBuffer) throws JshException {
        m_stdOutPrinter = stdOutPrinter;
        m_stdErrPrinter = stdErrPrinter;
        StdOutStream = JshStreamFactory.getInstance().makeStream("StdOut", _lock, JshStreamFactory.StreamLevel.TOP_LEVEL);
        StdErrStream = JshStreamFactory.getInstance().makeStream("StdErr", _lock, JshStreamFactory.StreamLevel.TOP_LEVEL);
        m_maxBuffer = maxBuffer;
        GsonBuilder jsonBuilder = new GsonBuilder();
        jsonBuilder.registerTypeAdapter(BasicStream.class, new BasicStreamSerializer());
        jsonBuilder.registerTypeAdapter(BasicTypeStream.class, new BasicTypeStreamSerializer());
        jsonBuilder.registerTypeAdapter(TopLevelStream.class, new TopLevelStreamSerializer());
        m_json = jsonBuilder.create();
    }

    /**
     * Increments the count of objects in the buffer.
     *
     * @param amount The number of objects that have been added to the buffer
     */
    void incrementCount(int amount) {
        synchronized (_lock) {
            m_bufferCount += amount;
            if (m_bufferCount >= m_maxBuffer) {
                flush();
            }
        }
    }

    /**
     * Writes all current buffered objects to the initialized stdout and stderr.
     */
    public void flush() {
        synchronized (_lock) {
            m_stdOutPrinter.println(m_json.toJson(StdOutStream.getSubStreams()));
            m_stdErrPrinter.println(m_json.toJson(StdErrStream.getSubStreams()));
            m_bufferCount = 0;
        }
    }

}
