package edu.wpi.cs.jsh;

import java.lang.reflect.Type;
import java.util.Collection;
import java.util.Map;

/**
 * This is an interface for the idea of a JSH style stream or substream. Streams have the following concepts:
 * <ul>
 * <li>You can output types on a specific key</li>
 * <li>You can create new substreams from this stream</li>
 * <li>All types output on a specific key must be the same</li>
 * </ul>
 * A stream that implements this interface should not directly output values. It outputs the values of the substreams,
 * which are either streams or substreams
 */
public interface Stream extends NamedStream {

    /**
     * Gets a raw type stream of the specified generic type. This output stream can be of any type, but you should not use
     * {@link edu.wpi.cs.jsh.Stream} as the type. If you need a new substream, use the {@link edu.wpi.cs.jsh.Stream#getSubStream(String)}
     * method instead.
     *
     * @param <T>        The type of the substream. If the stream name is already a stream of a different type, an exception will be thrown
     * @param streamName The name of the substream.
     * @param streamType The type of T parameter. Due to Java generic limitations, this needs to passed in to be available at runtime
     * @return The raw type stream with the given name and type
     * @throws JshException Thrown if the stream name is not unique and has conflicting types
     */
    <T> TypeStream<T> getTypeStream(String streamName, Class<T> streamType) throws JshException;

    /**
     * Gets a new Stream that allows for additional substreams. The stream name must be unique, ie this stream does not already
     * have any other Streams or TypeStreams that have the same name. Streams do not output directly: if you need to directly output
     * a type, use the {@link Stream#getTypeStream(String, Class)} method instead.
     *
     * @param streamName The name of the substream to create
     * @return The substream with the given name
     * @throws JshException Thrown if the stream name is not unique and has conflicting types
     */
    Stream getSubStream(String streamName) throws JshException;

    /**
     * Gets all currently created substreams.
     *
     * @return All substreams
     */
    Map<String, NamedStream> getSubStreams();
}
